# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from time import sleep
from typing import Optional, Dict, Tuple, List, Callable, Iterable, Union, Any
import re
import os
import sys
import errno
from gettext import gettext as _
import logging
import logging.handlers
import argparse
from ctypes import cdll, byref, create_string_buffer
import traceback
import fcntl
import struct
import socket
import array
import time

import cairo

from blueman.main.Config import Config
from blueman.main.DBusProxies import AppletService, DBusProxyFailed
from blueman.Constants import BIN_DIR, ICON_PATH

import gi
gi.require_version("Gtk", "3.0")
gi.require_version("Gdk", "3.0")
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import Gio

__all__ = ["check_bluetooth_status", "launch", "setup_icon_path", "adapter_path_to_name", "e_", "bmexit",
           "format_bytes", "create_menuitem", "have", "set_proc_title", "create_logger", "create_parser", "open_rfcomm",
           "get_local_interfaces"]


def check_bluetooth_status(message: str, exitfunc: Callable[[], Any]) -> None:
    try:
        applet = AppletService()
    except DBusProxyFailed as e:
        logging.exception(e)
        print("Blueman applet needs to be running")
        exitfunc()
        return

    if "PowerManager" not in applet.QueryPlugins():
        return

    if not applet.GetBluetoothStatus():
        d = Gtk.MessageDialog(
            type=Gtk.MessageType.ERROR, icon_name="blueman",
            text=_("Bluetooth Turned Off"), secondary_text=message)
        d.add_button(_("Exit"), Gtk.ResponseType.NO)
        d.add_button(_("Enable Bluetooth"), Gtk.ResponseType.YES)

        resp = d.run()
        d.destroy()

        if resp != Gtk.ResponseType.YES:
            exitfunc()
            return

    applet.SetBluetoothStatus('(b)', True)
    if not applet.GetBluetoothStatus():
        print('Failed to enable bluetooth')
        exitfunc()
        return

    config = Config("org.blueman.plugins.powermanager")
    if config["auto-power-on"] is None:
        d = Gtk.MessageDialog(
            type=Gtk.MessageType.QUESTION, icon_name="blueman",
            text=_("Shall bluetooth get enabled automatically?"))
        d.add_button(_("Yes"), Gtk.ResponseType.YES)
        d.add_button(_("No"), Gtk.ResponseType.NO)

        resp = d.run()
        d.destroy()

        config["auto-power-on"] = resp == Gtk.ResponseType.YES


def launch(
    cmd: str,
    paths: Optional[Iterable[str]] = None,
    system: bool = False,
    icon_name: Optional[str] = None,
    name: str = "blueman",
    sn: bool = True,
) -> bool:
    """Launch a gui app with starup notification"""
    context = None
    gtktimestamp = Gtk.get_current_event_time()
    if gtktimestamp == 0:
        logging.info("Gtk eventtime is 0, not using LaunchContext")
        timestamp = int(time.clock_gettime(time.CLOCK_MONOTONIC_RAW))
    else:
        timestamp = gtktimestamp
        display = Gdk.Display.get_default()
        assert display
        context = display.get_app_launch_context()
        context.set_timestamp(timestamp)

    if sn:
        flags = Gio.AppInfoCreateFlags.SUPPORTS_STARTUP_NOTIFICATION
    else:
        flags = Gio.AppInfoCreateFlags.NONE

    env = os.environ
    env["BLUEMAN_EVENT_TIME"] = str(timestamp)

    if not system:
        cmd = os.path.join(BIN_DIR, cmd)
    else:
        cmd = os.path.expanduser(cmd)

    if paths:
        files: Optional[List[Gio.File]] = [Gio.File.new_for_commandline_arg(p) for p in paths]
    else:
        files = None

    if icon_name and context is not None:
        context.set_icon_name(icon_name)

    appinfo = Gio.AppInfo.create_from_commandline(cmd, name, flags)
    launched: bool = appinfo.launch(files, context)

    if not launched:
        logging.error(f"Command: {cmd} failed")

    return launched


def setup_icon_path() -> None:
    ic = Gtk.IconTheme.get_default()
    ic.prepend_search_path(ICON_PATH)


def adapter_path_to_name(path: Optional[str]) -> Optional[str]:
    if path is None or path == '':
        return None

    match = re.search(r".*(hci[0-9]*)", path)
    if match:
        return match.group(1)
    return None


# format error
def e_(msg: Union[str, Exception]) -> Tuple[str, Optional[str]]:
    if isinstance(msg, Exception):
        return str(msg), traceback.format_exc()
    else:
        s = msg.strip().split(": ")[-1]
        return s, None


def format_bytes(size: float) -> Tuple[float, str]:
    size = float(size)
    if size < 1024:
        ret = size
        suffix = "B"
    elif 1024 < size < (1024 * 1024):
        ret = size / 1024
        suffix = "KB"
    elif (1024 * 1024) < size < (1024 * 1024 * 1024):
        ret = size / (1024 * 1024)
        suffix = "MB"
    else:
        ret = size / (1024 * 1024 * 1024)
        suffix = "GB"

    return ret, suffix


def create_menuitem(
    text: str,
    icon_name: Optional[str] = None,
    pixbuf: Optional[GdkPixbuf.Pixbuf] = None,
    surface: Optional[cairo.Surface] = None,
) -> Gtk.ImageMenuItem:
    image = Gtk.Image(pixel_size=16)
    if icon_name:
        image.set_from_icon_name(icon_name, Gtk.IconSize.MENU)
    elif surface:
        image.set_from_surface(surface)
    elif pixbuf:
        image.set_from_pixbuf(pixbuf)
    else:
        raise ValueError("At least provide one of, icon name, surface or pixbuf")

    item = Gtk.ImageMenuItem(label=text, image=image, use_underline=True)
    child = item.get_child()
    assert isinstance(child, Gtk.AccelLabel)
    child.set_use_markup(True)
    item.show_all()

    return item


def have(t: str) -> Optional[str]:
    paths = os.environ['PATH'] + ':/sbin:/usr/sbin'
    for path in paths.split(os.pathsep):
        exec_path = os.path.join(path, t)
        exists = os.path.exists(exec_path)
        executable = os.access(exec_path, os.EX_OK)
        if exists and executable:
            return exec_path
    return None


def set_proc_title(name: Optional[str] = None) -> int:
    """Set the process title"""

    if not name:
        name = os.path.basename(sys.argv[0])

    libc = cdll.LoadLibrary('libc.so.6')
    buff = create_string_buffer(len(name) + 1)
    buff.value = name.encode("UTF-8")
    ret: int = libc.prctl(15, byref(buff), 0, 0, 0)

    if ret != 0:
        logging.error("Failed to set process title")

    return ret


logger_format = '%(name)s %(asctime)s %(levelname)-8s %(module)s:%(lineno)s %(funcName)-10s: %(message)s'
syslog_logger_format = '%(name)s %(levelname)s %(module)s:%(lineno)s %(funcName)s: %(message)s'
logger_date_fmt = '%H.%M.%S'


def create_logger(
    log_level: int,
    name: str,
    log_format: Optional[str] = None,
    date_fmt: Optional[str] = None,
    syslog: bool = False,
) -> logging.Logger:
    if log_format is None:
        log_format = logger_format
    if date_fmt is None:
        date_fmt = logger_date_fmt
    logging.basicConfig(level=log_level, format=log_format, datefmt=date_fmt)

    logger = logging.getLogger(None)  # Root logger
    logger.name = name

    if syslog:
        syslog_handler = logging.handlers.SysLogHandler(address="/dev/log")
        syslog_formatter = logging.Formatter(syslog_logger_format)
        syslog_handler.setFormatter(syslog_formatter)
        logger.addHandler(syslog_handler)

    return logger


def create_parser(
    parser: Optional[argparse.ArgumentParser] = None,
    syslog: bool = True,
    loglevel: bool = True,
) -> argparse.ArgumentParser:
    if parser is None:
        parser = argparse.ArgumentParser()

    if loglevel:
        parser.add_argument("--loglevel", dest="LEVEL", default="warning")

    if syslog:
        parser.add_argument("--syslog", dest="syslog", action="store_true")

    return parser


def open_rfcomm(file: str, mode: int) -> int:
    try:
        return os.open(file, mode | os.O_EXCL | os.O_NONBLOCK | os.O_NOCTTY)
    except OSError as err:
        if err.errno == errno.EBUSY:
            logging.warning(f"{file} is busy, delaying 2 seconds")
            sleep(2)
            return open_rfcomm(file, mode)
        else:
            raise


def _netmask_for_ifacename(name: str, sock: socket.socket) -> Optional[str]:
    siocgifnetmask = 0x891b
    bytebuf = struct.pack('256s', name.encode('utf-8'))
    try:
        ret = fcntl.ioctl(sock.fileno(), siocgifnetmask, bytebuf)
    except OSError:
        logging.error('siocgifnetmask failed')
        return None

    return socket.inet_ntoa(ret[20:24])


def get_local_interfaces() -> Dict[str, Tuple[str, Optional[str]]]:
    """ Returns a dictionary of name:ip, mask key value pairs. """
    siocgifconf = 0x8912
    names = array.array('B', 4096 * b'\0')
    names_address, names_length = names.buffer_info()
    mutable_byte_buffer = struct.pack('iL', 4096, names_address)

    try:
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
            try:
                mutated_byte_buffer = fcntl.ioctl(sock.fileno(), siocgifconf, mutable_byte_buffer)
            except OSError:
                logging.error('siocgifconf failed')
                return {}

            max_bytes_out, names_address_out = struct.unpack('iL', mutated_byte_buffer)
            namestr = names.tobytes()

            ip_dict = {}
            for i in range(0, max_bytes_out, 40):
                name = namestr[i: i + 16].split(b'\0', 1)[0].decode('utf-8')
                ipaddr = socket.inet_ntoa(namestr[i + 20: i + 24])
                mask = _netmask_for_ifacename(name, sock)
                ip_dict[name] = (ipaddr, mask)
    except OSError:
        logging.error('Socket creation failed', exc_info=True)
        return {}

    return ip_dict


def bmexit(msg: Optional[Union[str, int]] = None) -> None:
    raise SystemExit(msg)



from typing import TypeVar, Type

import gi
from blueman.Constants import UI_PATH

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk


class Builder(Gtk.Builder):
    def __init__(self, filename: str):
        super().__init__(translation_domain="blueman")
        self.add_from_file(UI_PATH + "/" + filename)

    T = TypeVar("T", bound=Gtk.Widget)

    def get_widget(self, name: str, widget_type: Type[T]) -> T:
        widget = self.get_object(name)
        assert isinstance(widget, widget_type)
        return widget



from typing import List, Optional, Collection, Iterable

import cairo
import gi

from blueman.bluemantyping import GSignals

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
gi.require_version("Gdk", "3.0")
from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import GLib


class AnimBase(GObject.GObject):
    __gsignals__: GSignals = {
        'animation-finished': (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, state: float = 1.0) -> None:
        super().__init__()
        self._source: Optional[int] = None
        self._state = state
        self.frozen = False
        self.fps = 24.0

    def _do_transition(self) -> bool:
        if abs(self._end - self._start) < 0.000001:
            return False

        self._state += self._step_size

        if self._end - self._start < 0:
            if self._state <= self._end:
                self._state = self._end
                self._state_changed(self._state)
                self._source = None
                self.emit("animation-finished")
                return False
        else:
            if self._state >= self._end:
                self._state = self._end
                self._state_changed(self._state)
                self._source = None
                self.emit("animation-finished")
                return False

        self._state_changed(self._state)
        return True

    def thaw(self) -> None:
        self.frozen = False

    def freeze(self) -> None:
        self.frozen = True

    def animate(self, start: float = 1.0, end: float = 0.0, duration: int = 1000) -> None:
        if self.frozen:
            self.emit("animation-finished")
            return

        self._state = start
        self._start = start
        self._end = end
        self._duration = duration

        if self._source:
            GLib.source_remove(self._source)

        try:
            self._step_size = (end - start) / (self.fps * (duration / 1000.0))
        except ZeroDivisionError:
            self._state = end

            return

        self._state_changed(self._state)
        self._source = GLib.timeout_add(int(1.0 / self.fps * 1000), self._do_transition)

    def _state_changed(self, state: float) -> None:
        self.state_changed(state)

    def state_changed(self, state: float) -> None:
        pass

    def get_state(self) -> float:
        return self._state

    def set_state(self, state: float) -> None:
        self._state = state
        self._state_changed(state)

    def is_animating(self) -> bool:
        return self._source is not None


class TreeRowFade(AnimBase):
    def __init__(self, tw: Gtk.TreeView, path: Gtk.TreePath, columns: Optional[Collection[Gtk.TreeViewColumn]] = None
                 ) -> None:
        super().__init__(1.0)
        self.tw = tw

        self.sig: Optional[int] = self.tw.connect_after("draw", self.on_draw)

        assert tw.props.model is not None
        self.row = Gtk.TreeRowReference.new(tw.props.model, path)
        self.stylecontext = tw.get_style_context()
        self.columns = columns

    def unref(self) -> None:
        if self.sig is not None:
            self.tw.disconnect(self.sig)
            self.sig = None

    def get_iter(self) -> Optional[Gtk.TreeIter]:
        assert isinstance(self.tw.props.model, Gtk.TreeModel)
        path = self.row.get_path()
        assert path is not None
        return self.tw.props.model.get_iter(path)

    def on_draw(self, widget: Gtk.Widget, cr: cairo.Context) -> bool:
        if self.frozen:
            return False

        if not self.row.valid():
            if self.sig is not None:
                self.tw.disconnect(self.sig)
                self.sig = None
            return False

        path = self.row.get_path()

        color = self.stylecontext.get_background_color(Gtk.StateFlags.NORMAL)

        if not self.columns:
            self.columns = self.tw.get_columns()
        assert self.columns is not None

        for col in self.columns:
            rect = self.tw.get_background_area(path, col)
            cr.rectangle(rect.x, rect.y, rect.width, rect.height)

        cr.clip()

        cr.set_source_rgba(color.red, color.green, color.blue, 1.0 - self.get_state())
        cr.set_operator(cairo.OPERATOR_OVER)
        cr.paint()

        return False

    def state_changed(self, state: float) -> None:
        self.tw.queue_draw()


class CellFade(AnimBase):
    def __init__(self, tw: Gtk.TreeView, path: Gtk.TreePath, columns: Iterable[int]) -> None:
        super().__init__(1.0)
        self.tw = tw

        self.frozen = False
        self.sig: Optional[int] = tw.connect_after("draw", self.on_draw)
        assert tw.props.model is not None
        self.row = Gtk.TreeRowReference.new(tw.props.model, path)
        self.selection = tw.get_selection()
        self.columns: List[Optional[Gtk.TreeViewColumn]] = []
        for i in columns:
            self.columns.append(self.tw.get_column(i))

    def unref(self) -> None:
        if self.sig is not None:
            self.tw.disconnect(self.sig)
            self.sig = None

    def get_iter(self) -> Optional[Gtk.TreeIter]:
        assert isinstance(self.tw.props.model, Gtk.TreeModel)
        path = self.row.get_path()
        assert path is not None
        return self.tw.props.model.get_iter(path)

    def on_draw(self, _widget: Gtk.Widget, cr: cairo.Context) -> bool:
        if self.frozen:
            return False

        if not self.row.valid():
            if self.sig is not None:
                self.tw.disconnect(self.sig)
                self.sig = None

        path = self.row.get_path()

        # FIXME Use Gtk.render_background to render background.
        # However it does not use the correct colors/gradient.
        for col in self.columns:
            bg_rect = self.tw.get_background_area(path, col)
            rect = self.tw.get_cell_area(path, col)
            rect.y = bg_rect.y
            rect.height = bg_rect.height

            cr.rectangle(rect.x, rect.y, rect.width, rect.height)

        cr.clip()

        assert self.tw.props.model is not None
        maybe_selected = self.selection.get_selected()[1]
        if maybe_selected is not None:
            selected = self.tw.props.model.get_path(maybe_selected) == path
        else:
            selected = False

        stylecontext = self.tw.get_style_context()

        if selected:
            bg_color = stylecontext.get_background_color(Gtk.StateFlags.SELECTED)
        else:
            bg_color = stylecontext.get_background_color(Gtk.StateFlags.NORMAL)

        cr.set_source_rgb(bg_color.red, bg_color.green, bg_color.blue)
        cr.paint_with_alpha(1.0 - self.get_state())

        return False

    def state_changed(self, state: float) -> None:
        self.tw.queue_draw()


class WidgetFade(AnimBase):
    def __init__(self, widget: Gtk.Widget, color: Gdk.RGBA) -> None:
        super().__init__(1.0)

        self.widget = widget
        self.color = color

        self.sig = widget.connect_after("draw", self.on_draw)

    def on_draw(self, _widget: Gtk.Widget, cr: cairo.Context) -> bool:
        if not self.frozen:
            cr.set_source_rgba(self.color.red, self.color.green, self.color.blue, self.color.alpha - self.get_state())
            cr.set_operator(cairo.OPERATOR_OVER)
            cr.paint()
        return False

    def state_changed(self, state: float) -> None:
        self.widget.queue_draw()



import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

from blueman.bluez.Device import Device
from blueman.main.Config import Config


class AutoConnectConfig(Config):
    def __init__(self) -> None:
        super().__init__("org.blueman.plugins.autoconnect")

    def bind_to_menuitem(self, item: Gtk.CheckMenuItem, device: Device, uuid: str) -> None:
        data = device.get_object_path(), uuid

        def switch(active: bool) -> None:
            services = set(self["services"])
            if active:
                self["services"] = set(services).union({data})
            else:
                self["services"] = set(self["services"]).difference({data})

        def on_change(config: AutoConnectConfig, key: str) -> None:
            if key == "services":
                item.props.active = data in set(config[key])

        item.props.active = data in set(self["services"])
        item.connect("toggled", lambda i: switch(i.props.active))
        self.connect("changed", on_change)



from blueman.services.meta import NetworkService
from blueman.Sdp import NAP_SVCLASS_ID


class NetworkAccessPoint(NetworkService):
    __svclass_id__ = NAP_SVCLASS_ID
    __icon__ = "network-wireless"
    __priority__ = 81



from gettext import gettext as _
import logging
from typing import TYPE_CHECKING, Callable, Tuple, Optional

import gi

from blueman.bluez.Adapter import Adapter
from blueman.bluez.Device import Device
from blueman.gui.manager.ManagerDeviceList import ManagerDeviceList

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

if TYPE_CHECKING:
    from blueman.main.Manager import Blueman


class ManagerToolbar:
    def __init__(self, blueman: "Blueman") -> None:
        self.blueman = blueman

        self.blueman.List.connect("device-selected", self.on_device_selected)
        self.blueman.List.connect("device-property-changed", self.on_device_propery_changed)
        self.blueman.List.connect("adapter-changed", self.on_adapter_changed)
        self.blueman.List.connect("adapter-property-changed", self.on_adapter_property_changed)

        self.b_search = blueman.builder.get_widget("b_search", Gtk.ToolButton)
        self.b_search.connect("clicked", lambda button: blueman.inquiry())

        self.b_bond = blueman.builder.get_widget("b_bond", Gtk.ToolButton)
        self.b_bond.connect("clicked", self.on_action, self.blueman.bond)

        self.b_trust = blueman.builder.get_widget("b_trust", Gtk.ToolButton)
        self.b_trust.connect("clicked", self.on_action, self.blueman.toggle_trust)
        self.b_trust.set_homogeneous(False)

        self.b_trust.props.label = _("Untrust")
        (size, nsize) = Gtk.Widget.get_preferred_size(self.b_trust)
        self.b_trust.props.label = _("Trust")
        (size2, nsize2) = Gtk.Widget.get_preferred_size(self.b_trust)

        self.b_trust.props.width_request = max(size.width, size2.width)

        self.b_remove = blueman.builder.get_widget("b_remove", Gtk.ToolButton)
        self.b_remove.connect("clicked", self.on_action, self.blueman.remove)

        self.b_send = blueman.builder.get_widget("b_send", Gtk.ToolButton)
        self.b_send.props.sensitive = False
        self.b_send.connect("clicked", self.on_action, self.blueman.send)
        self.b_send.set_homogeneous(False)

        self.on_adapter_changed(blueman.List, blueman.List.get_adapter_path())

    def on_action(self, _button: Gtk.ToolButton, func: Callable[[Device], None]) -> None:
        device = self.blueman.List.get_selected_device()
        if device is not None:
            func(device)

    def on_adapter_property_changed(self, _lst: ManagerDeviceList, _adapter: Adapter,
                                    key_value: Tuple[str, object]) -> None:
        key, value = key_value
        if key == "Discovering":
            if value:
                self.b_search.props.sensitive = False
            else:
                self.b_search.props.sensitive = True

    def on_adapter_changed(self, _lst: ManagerDeviceList, adapter_path: Optional[str]) -> None:
        logging.debug(f"toolbar adapter {adapter_path}")
        if adapter_path is None:
            self.b_search.props.sensitive = False
            self.b_send.props.sensitive = False
        else:
            self.b_search.props.sensitive = True

    def on_device_selected(self, dev_list: ManagerDeviceList, device: Device, tree_iter: Gtk.TreeIter) -> None:
        if device is None or tree_iter is None:
            self.b_bond.props.sensitive = False
            self.b_remove.props.sensitive = False
            self.b_trust.props.sensitive = False
        else:
            row = dev_list.get(tree_iter, "paired", "trusted", "objpush")
            self.b_remove.props.sensitive = True
            if row["paired"]:
                self.b_bond.props.sensitive = False
            else:
                self.b_bond.props.sensitive = True

            if row["trusted"]:
                image = Gtk.Image(icon_name="blueman-untrust", pixel_size=24, visible=True)
                self.b_trust.props.icon_widget = image
                self.b_trust.props.sensitive = True
                self.b_trust.props.label = _("Untrust")

            else:
                image = Gtk.Image(icon_name="blueman-trust", pixel_size=24, visible=True)
                self.b_trust.props.icon_widget = image
                self.b_trust.props.sensitive = True
                self.b_trust.props.label = _("Trust")

            if row["objpush"]:
                self.b_send.props.sensitive = True
            else:
                self.b_send.props.sensitive = False

    def on_device_propery_changed(self, dev_list: ManagerDeviceList, device: Device, tree_iter: Gtk.TreeIter,
                                  key_value: Tuple[str, object]) -> None:
        key, value = key_value
        if dev_list.compare(tree_iter, dev_list.selected()):
            if key == "Trusted" or key == "Paired" or key == "UUIDs":
                self.on_device_selected(dev_list, device, tree_iter)



from typing import Callable, Iterable, TYPE_CHECKING, overload, Any, cast, Mapping, Optional

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from blueman.Functions import create_menuitem

if TYPE_CHECKING:
    from typing_extensions import Protocol

    from blueman.plugins.applet.Menu import MenuItemDict, SubmenuItemDict

    class MenuItemActivator(Protocol):
        @overload
        def __call__(self, idx: int) -> None:
            ...

        @overload
        def __call__(self, idx: int, subid: int) -> None:
            ...


@overload
def build_menu(items: Iterable["MenuItemDict"], activate: "MenuItemActivator") -> Gtk.Menu:
    ...


@overload
def build_menu(items: Iterable["SubmenuItemDict"], activate: Callable[[int], None]) -> Gtk.Menu:
    ...


def build_menu(items: Iterable[Mapping[str, Any]], activate: Callable[..., None]) -> Gtk.Menu:
    menu = Gtk.Menu()
    for index, item in enumerate(items):
        if 'text' in item and 'icon_name' in item:
            gtk_item: Gtk.MenuItem = create_menuitem(item['text'], item['icon_name'])
            label = gtk_item.get_child()
            assert isinstance(label, Gtk.Label)
            if item['markup']:
                label.set_markup_with_mnemonic(item['text'])
            else:
                label.set_text_with_mnemonic(item['text'])
            gtk_item.connect('activate', cast(Callable[[Gtk.MenuItem], None], lambda _, idx=index: activate(idx)))
            if 'submenu' in item:
                gtk_item.set_submenu(build_menu(item['submenu'], cast(Callable[[int], None],
                                                                      lambda subid, idx=index: activate(idx, subid))))
            if 'tooltip' in item:
                gtk_item.props.tooltip_text = item['tooltip']
            gtk_item.props.sensitive = item['sensitive']
        else:
            gtk_item = Gtk.SeparatorMenuItem()
        gtk_item.show()
        menu.append(gtk_item)
    return menu


class GtkStatusIcon:
    def __init__(self, icon_name: str, on_activate_menu_item: "MenuItemActivator",
                 on_activate_status_icon: Callable[[], None]) -> None:
        self._on_activate = on_activate_menu_item
        self.indicator = Gtk.StatusIcon(icon_name=icon_name)
        self.indicator.set_title('blueman')
        self.indicator.connect('popup-menu', self.on_popup_menu)
        self.indicator.connect('activate', lambda _status_icon: on_activate_status_icon())
        self._menu: Optional[Gtk.Menu] = None

    def on_popup_menu(self, _status_icon: Gtk.StatusIcon, _button: int, _activate_time: int) -> None:
        if self._menu:
            self._menu.popup_at_pointer(None)

    def set_icon(self, icon_name: str) -> None:
        self.indicator.props.icon_name = icon_name

    def set_text(self, text: str) -> None:
        self.indicator.props.tooltip_markup = text

    def set_visibility(self, visible: bool) -> None:
        self.indicator.props.visible = visible

    def set_menu(self, menu: Iterable["MenuItemDict"]) -> None:
        self._menu = build_menu(menu, self._on_activate)



from html import escape
from typing import Optional, Any, List

from blueman.bluez.Device import Device
from blueman.gui.DeviceList import DeviceList
from gi.repository import Pango

import gi

from blueman.gui.GenericList import ListDataDict

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk


class DeviceSelectorList(DeviceList):
    def __init__(self, adapter_name: Optional[str] = None) -> None:
        tabledata: List[ListDataDict] = [
            # device picture
            {"id": "device_icon", "type": str, "renderer": Gtk.CellRendererPixbuf(stock_size=Gtk.IconSize.MENU),
             "render_attrs": {"icon_name": 0}},
            # device caption
            {"id": "caption", "type": str, "renderer": Gtk.CellRendererText(ellipsize=Pango.EllipsizeMode.END),
             "render_attrs": {"markup": 1},
             "view_props": {"expand": True}},
            {"id": "paired_icon", "type": str, "renderer": Gtk.CellRendererPixbuf(stock_size=Gtk.IconSize.MENU),
             "render_attrs": {"icon_name": 2}},
            {"id": "trusted_icon", "type": str, "renderer": Gtk.CellRendererPixbuf(stock_size=Gtk.IconSize.MENU),
             "render_attrs": {"icon_name": 3}}
        ]

        super().__init__(adapter_name, tabledata, headers_visible=False)

    def on_icon_theme_changed(self, _icon_them: Gtk.IconTheme) -> None:
        for row in self.liststore:
            device = self.get(row.iter, "device")["device"]
            self.row_setup_event(row.iter, device)

    def row_setup_event(self, tree_iter: Gtk.TreeIter, device: Device) -> None:
        self.row_update_event(tree_iter, "Trusted", device['Trusted'])
        self.row_update_event(tree_iter, "Paired", device['Paired'])
        self.row_update_event(tree_iter, "Alias", device['Alias'])
        self.row_update_event(tree_iter, "Icon", device['Icon'])

    def row_update_event(self, tree_iter: Gtk.TreeIter, key: str, value: Any) -> None:
        if key == "Trusted":
            if value:
                self.set(tree_iter, trusted_icon="blueman-trust")
            else:
                self.set(tree_iter, trusted_icon=None)

        elif key == "Paired":
            if value:
                self.set(tree_iter, paired_icon="dialog-password")
            else:
                self.set(tree_iter, paired_icon=None)

        elif key == "Alias":
            self.set(tree_iter, caption=escape(value))

        elif key == "Icon":
            self.set(tree_iter, device_icon=value)



import importlib
import logging
import os
from typing import Optional

import blueman.plugins.mechanism
from blueman.Constants import POLKIT
from gi.repository import GLib, Gio

from blueman.main.DbusService import DbusService, DbusError
from blueman.plugins.MechanismPlugin import MechanismPlugin


class Timer:
    def __init__(self, loop: GLib.MainLoop):
        self.time = 0
        self.stopped = False
        self._loop = loop
        GLib.timeout_add(1000, self.tick)

    def tick(self) -> bool:
        if not self.stopped:
            self.time += 1
            if self.time == (9999 if 'BLUEMAN_SOURCE' in os.environ else 30):
                logging.info("Exiting")
                self._loop.quit()

        return True

    def reset(self) -> None:
        self.time = 0

    def stop(self) -> None:
        self.stopped = True

    def resume(self) -> None:
        self.stopped = False
        self.reset()


class MechanismApplication(DbusService):
    def __init__(self, stoptimer: bool):
        super().__init__("org.blueman.Mechanism", "org.blueman.Mechanism", "/org/blueman/mechanism", Gio.BusType.SYSTEM)
        self._loop = GLib.MainLoop()
        self.timer = Timer(self._loop)
        if stoptimer:
            self.timer.stop()

        if POLKIT:
            try:
                self.pk: Optional[Gio.DBusProxy] = Gio.DBusProxy.new_for_bus_sync(
                    Gio.BusType.SYSTEM,
                    Gio.DBusProxyFlags.NONE,
                    None,
                    'org.freedesktop.PolicyKit1',
                    '/org/freedesktop/PolicyKit1/Authority',
                    'org.freedesktop.PolicyKit1.Authority')
            except Exception as e:
                logging.exception(e)
                self.pk = None
        else:
            self.pk = None

        path = os.path.dirname(blueman.plugins.mechanism.__file__)
        plugins = []
        for root, dirs, files in os.walk(path):
            for f in files:
                if f.endswith(".py") and not (f.endswith(".pyc") or f.endswith("_.py")):
                    plugins.append(f[0:-3])

        for plugin in plugins:
            try:
                importlib.import_module(f"blueman.plugins.mechanism.{plugin}")
            except ImportError:
                logging.error(f"Skipping plugin {plugin}", exc_info=True)

        classes = MechanismPlugin.__subclasses__()
        for cls in classes:
            logging.info(f"loading {cls.__name__}")
            cls(self)

        self.register()

    def run(self) -> None:
        self._loop.run()

    def confirm_authorization(self, subject: str, action_id: str) -> None:
        self.timer.reset()
        if not POLKIT:
            return
        else:
            if not self.pk:
                raise DbusError("Blueman was built with PolicyKit-1 support, but it's not available on the system")

        v_subject = GLib.Variant('s', subject)
        res = self.pk.CheckAuthorization('((sa{sv})sa{ss}us)', ("system-bus-name", {"name": v_subject}),
                                         action_id, {}, 1, "")

        logging.debug(str(res))
        (is_authorized, is_challenge, details) = res
        if not is_authorized:
            raise DbusError("Not authorized")



import os.path
import logging
from gettext import gettext as _
from html import escape
import random
from xml.etree import ElementTree
from typing import Dict, Optional, overload, Callable, Union, TYPE_CHECKING, Tuple, Any, List

from blueman.bluez.Device import Device
from blueman.bluez.AgentManager import AgentManager
from blueman.Sdp import ServiceUUID
from blueman.Constants import PKGDATA_DIR
from blueman.gui.Notification import Notification, _NotificationBubble, _NotificationDialog
from blueman.main.Builder import Builder
from blueman.main.DbusService import DbusService, DbusError

from gi.repository import Gio

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

if TYPE_CHECKING:
    from typing_extensions import Literal


def bt_class_to_string(bt_class: int) -> Optional[str]:
    n1 = (bt_class & 0x1f00) >> 8
    if n1 == 0x03:
        return "network"
    elif n1 == 0x04:
        n2 = (bt_class & 0xfc) >> 2
        if n2 in (0x01, 0x02):
            return "headset"
        elif n2 == 0x06:
            return "headphone"
        else:
            return "audio"
    elif n1 == 0x05:
        n2 = (bt_class & 0xc0) >> 6
        if n2 == 0x00:
            n3 = (bt_class & 0x1e) >> 2
            if n3 in [0x01, 0x02]:
                return "joypad"
        elif n2 == 0x01:
            return "keyboard"
        elif n2 == 0x02:
            n3 = (bt_class & 0x1e) >> 2
            if n3 == 0x05:
                return "tablet"
            else:
                return "mouse"
    elif n1 == 0x06:
        if bt_class & 0x80:
            return "printer"
    return None


PIN_SEARCHES = [
    "./device[@oui='{oui}'][@type='{type}'][@name='{name}']",
    "./device[@oui='{oui}'][@type='{type}']",
    "./device[@oui='{oui}'][@name='{name}']",
    "./device[@type='{type}'][@name='{name}']",
    "./device[@oui='{oui}']",
    "./device[@name='{name}']",
    "./device[@type='{type}']",
]


class BluezErrorCanceled(DbusError):
    _name = "org.bluez.Error.Canceled"


class BluezErrorRejected(DbusError):
    _name = "org.bluez.Error.Rejected"


class BluezAgent(DbusService):
    __agent_path = '/org/bluez/agent/blueman'

    def __init__(self) -> None:
        super().__init__(None, "org.bluez.Agent1", self.__agent_path, Gio.BusType.SYSTEM)

        self.add_method("Release", (), "", self._on_release)
        self.add_method("RequestPinCode", ("o",), "s", self._on_request_pin_code, is_async=True)
        self.add_method("DisplayPinCode", ("o", "s"), "", self._on_display_pin_code)
        self.add_method("RequestPasskey", ("o",), "u", self._on_request_passkey, is_async=True)
        self.add_method("DisplayPasskey", ("o", "u", "q"), "", self._on_display_passkey)
        self.add_method("RequestConfirmation", ("o", "u"), "", self._on_request_confirmation, is_async=True)
        self.add_method("RequestAuthorization", ("o",), "", self._on_request_authorization, is_async=True)
        self.add_method("AuthorizeService", ("o", "s"), "", self._on_authorize_service, is_async=True)
        self.add_method("Cancel", (), "", self._on_cancel)

        self.dialog: Optional[Gtk.Dialog] = None
        self._db: Optional[ElementTree.ElementTree] = None
        self._devhandlerids: Dict[str, int] = {}
        self._notification: Optional[Union[_NotificationBubble, _NotificationDialog]] = None
        self._service_notifications: List[Union[_NotificationBubble, _NotificationDialog]] = []

    def register_agent(self) -> None:
        logging.info("Register Agent")
        self.register()
        AgentManager().register_agent(self.__agent_path, "KeyboardDisplay", default=True)

    def unregister_agent(self) -> None:
        logging.info("Unregister Agent")
        self.unregister()
        AgentManager().unregister_agent(self.__agent_path)

    def build_passkey_dialog(self, device_alias: str, dialog_msg: str, is_numeric: bool
                             ) -> Tuple[Gtk.Dialog, Gtk.Entry]:
        def on_insert_text(editable: Gtk.Entry, new_text: str, _new_text_length: int, _position: int) -> None:
            if not new_text.isdigit():
                editable.stop_emission("insert-text")

        builder = Builder("applet-passkey.ui")

        dialog = builder.get_widget("dialog", Gtk.Dialog)

        dialog.props.icon_name = "blueman"
        dev_name = builder.get_widget("device_name", Gtk.Label)
        dev_name.set_markup(device_alias)
        msg = builder.get_widget("message", Gtk.Label)
        msg.set_text(dialog_msg)
        pin_entry = builder.get_widget("pin_entry", Gtk.Entry)
        show_input = builder.get_widget("show_input_check", Gtk.CheckButton)
        if is_numeric:
            pin_entry.set_max_length(6)
            pin_entry.set_width_chars(6)
            pin_entry.connect("insert-text", on_insert_text)
            show_input.hide()
        else:
            pin_entry.set_max_length(16)
            pin_entry.set_width_chars(16)
            pin_entry.set_visibility(False)
        show_input.connect("toggled", lambda x: pin_entry.set_visibility(x.props.active))
        accept_button = builder.get_widget("accept", Gtk.Button)
        pin_entry.connect("changed", lambda x: accept_button.set_sensitive(x.get_text() != ''))

        return dialog, pin_entry

    def get_device_string(self, device_path: str) -> str:
        device = Device(obj_path=device_path)
        return f"<b>{escape(device['Alias'])}</b> ({device['Address']})"

    def _lookup_default_pin(self, device_path: str) -> Optional[str]:
        if not self._db:
            self._db = ElementTree.parse(os.path.join(PKGDATA_DIR, 'pin-code-database.xml'))

        device = Device(obj_path=device_path)
        lookup_dict = {
            'name': device['Name'],
            'type': bt_class_to_string(device['Class']),
            'oui': device['Address'][:9]
        }

        pin = None
        for s in PIN_SEARCHES:
            search = s.format(**lookup_dict)
            entry = self._db.find(search)
            if entry is not None:
                pin = entry.get('pin')
                break

        if pin is not None:
            if 'max:' in pin:
                pin = "".join(random.sample('123456789', int(pin[-1])))
        return pin

    @overload
    def ask_passkey(self, dialog_msg: str, is_numeric: "Literal[True]", device_path: str, ok: Callable[[int], None],
                    err: Callable[[Union[BluezErrorCanceled, BluezErrorRejected]], None]) -> None:
        ...

    @overload
    def ask_passkey(self, dialog_msg: str, is_numeric: "Literal[False]", device_path: str, ok: Callable[[str], None],
                    err: Callable[[Union[BluezErrorCanceled, BluezErrorRejected]], None]) -> None:
        ...

    def ask_passkey(self, dialog_msg: str, is_numeric: bool, device_path: str, ok: Callable[[Any], None],
                    err: Callable[[Union[BluezErrorCanceled, BluezErrorRejected]], None]) -> None:
        def passkey_dialog_cb(dialog: Gtk.Dialog, response_id: int) -> None:
            if response_id == Gtk.ResponseType.ACCEPT:
                ret = pin_entry.get_text()
                ok(int(ret) if is_numeric else ret)
            else:
                err(BluezErrorRejected("Rejected"))
            dialog.destroy()
            self.dialog = None

        dev_str = self.get_device_string(device_path)
        notify_message = _("Pairing request for %s") % dev_str

        if self.dialog:
            logging.info("Agent: Another dialog still active, cancelling")
            err(BluezErrorCanceled("Canceled"))

        self.dialog, pin_entry = self.build_passkey_dialog(dev_str, dialog_msg, is_numeric)
        if not self.dialog:
            logging.error("Agent: Failed to build dialog")
            err(BluezErrorCanceled("Canceled"))

        Notification(_("Bluetooth Authentication"), notify_message, icon_name="blueman").show()

        self.dialog.connect("response", passkey_dialog_cb)
        self.dialog.present()

    # Workaround BlueZ not calling the Cancel method, see #164
    def _on_device_property_changed(self, device: Device, key: str, value: Any, path: str) -> None:
        if (key == "Paired" and value) or (key == "Connected" and not value):
            handlerid = self._devhandlerids.pop(path)
            device.disconnect_signal(handlerid)
            self._on_cancel()

    def _on_release(self) -> None:
        logging.info("Agent.Release")
        self._on_cancel()
        self.unregister()

    def _on_cancel(self) -> None:
        logging.info("Agent.Cancel")
        if self.dialog:
            self.dialog.response(Gtk.ResponseType.REJECT)
        self._close()

    def _close(self) -> None:
        if self._notification is not None:
            self._notification.close()
            self._notification = None

    def _on_request_pin_code(self, device_path: str, ok: Callable[[str], None],
                             err: Callable[[Union[BluezErrorCanceled, BluezErrorRejected]], None]) -> None:
        logging.info("Agent.RequestPinCode")
        dialog_msg = _("Enter PIN code for authentication:")

        default_pin = self._lookup_default_pin(device_path)
        if default_pin is not None:
            logging.info(f"Sending default pin: {default_pin}")
            ok(default_pin)
            return

        self.ask_passkey(dialog_msg, False, device_path, ok, err)
        if self.dialog:
            self.dialog.present()

    def _on_request_passkey(self, device: str, ok: Callable[[int], None],
                            err: Callable[[Union[BluezErrorCanceled, BluezErrorRejected]], None]) -> None:
        logging.info("Agent.RequestPasskey")
        dialog_msg = _("Enter passkey for authentication:")
        self.ask_passkey(dialog_msg, True, device, ok, err)
        if self.dialog:
            self.dialog.present()

    def _on_display_passkey(self, device: str, passkey: int, entered: int) -> None:
        logging.info(f"DisplayPasskey ({device}, {passkey:d} {entered:d})")
        dev = Device(obj_path=device)
        self._devhandlerids[device] = dev.connect_signal("property-changed", self._on_device_property_changed)

        key = f"{passkey:06}"
        notify_message = _("Pairing passkey for") + f" {self.get_device_string(device)}: " \
                                                    f"{key[:entered]}<b>{key[entered]}</b>{key[entered+1:]}"
        self._close()
        self._notification = Notification("Bluetooth", notify_message, 0, icon_name="blueman")
        self._notification.show()

    def _on_display_pin_code(self, device: str, pin_code: str) -> None:
        logging.info(f'DisplayPinCode ({device}, {pin_code})')
        dev = Device(obj_path=device)
        self._devhandlerids[device] = dev.connect_signal("property-changed", self._on_device_property_changed)

        notify_message = _("Pairing PIN code for") + f" {self.get_device_string(device)}: {pin_code}"
        self._notification = Notification("Bluetooth", notify_message, 0, icon_name="blueman")
        self._notification.show()

    def _on_request_confirmation(self, device_path: str, passkey: Optional[int], ok: Callable[[], None],
                                 err: Callable[[BluezErrorCanceled], None]) -> None:
        def on_confirm_action(action: str) -> None:
            if action == "confirm":
                ok()
            else:
                err(BluezErrorCanceled("User canceled pairing"))

        logging.info("Agent.RequestConfirmation")
        notify_message = _("Pairing request for:") + f"\n{self.get_device_string(device_path)}"

        if passkey:
            notify_message += "\n" + _("Confirm value for authentication:") + f" <b>{passkey:06}</b>"
        actions = [("confirm", _("Confirm")), ("deny", _("Deny"))]

        self._notification = Notification("Bluetooth", notify_message, 0, actions, on_confirm_action,
                                          icon_name="blueman")
        self._notification.show()

    def _on_request_authorization(self, device: str, ok: Callable[[], None],
                                  err: Callable[[BluezErrorCanceled], None]) -> None:
        self._on_request_confirmation(device, None, ok, err)

    def _on_authorize_service(self, device: str, uuid: str, ok: Callable[[], None],
                              err: Callable[[BluezErrorRejected], None]) -> None:
        def on_auth_action(action: str) -> None:
            logging.info(action)

            if action == "always":
                Device(obj_path=device).set("Trusted", True)
            if action == "always" or action == "accept":
                ok()
            else:
                err(BluezErrorRejected("Rejected"))

            self._service_notifications.remove(n)

        logging.info("Agent.Authorize")
        dev_str = self.get_device_string(device)
        service = ServiceUUID(uuid).name
        notify_message = \
            _("Authorization request for:") + f"\n{dev_str}\n" + _("Service:") + f" <b>{service}</b>"
        actions = [("always", _("Always accept")),
                   ("accept", _("Accept")),
                   ("deny", _("Deny"))]

        n = Notification(_("Bluetooth Authentication"), notify_message, 0, actions, on_auth_action, icon_name="blueman")
        n.show()
        self._service_notifications.append(n)



import logging
import sys
import traceback
from typing import Dict, Tuple, Callable, Set, Optional, Any, Collection

from gi.repository import Gio, GLib


class DbusError(Exception):
    _name = "org.blueman.Error"

    def __init__(self, message: str) -> None:
        self._message = message

    @property
    def name(self) -> str:
        return self._name

    @property
    def message(self) -> str:
        return self._message


class DbusService:
    def __init__(self, bus_name: Optional[str], interface_name: str, path: str, bus_type: Gio.BusType) -> None:
        self._bus = Gio.bus_get_sync(bus_type)
        if bus_name:
            Gio.bus_own_name(bus_type, bus_name, Gio.BusNameOwnerFlags.NONE, None, None, None)
        self._methods: Dict[str, Tuple[Tuple[str, ...], str, Callable[..., None], Set[str]]] = {}
        self._signals: Dict[str, str] = {}
        self._interface_name = interface_name
        self._path = path
        self._regid: Optional[int] = None

    def add_method(self, name: str, arguments: Tuple[str, ...], return_value: str, method: Callable[..., None],
                   pass_sender: bool = False, is_async: bool = False) -> None:
        if name in self._signals:
            raise Exception(f"{name} already defined")

        options = set()
        if pass_sender:
            options.add("sender")
        if is_async:
            options.add("async")
        self._methods[name] = (arguments, return_value, method, options)
        self._reregister()

    def remove_method(self, name: str) -> None:
        del self._methods[name]
        self._reregister()

    def add_signal(self, name: str, signature: str) -> None:
        if name in self._signals:
            raise Exception(f"{name} already defined")

        self._signals[name] = signature
        self._reregister()

    def remove_signal(self, name: str) -> None:
        del self._signals[name]
        self._reregister()

    def emit_signal(self, name: str, *args: Any) -> None:
        self._bus.emit_signal(None, self._path, self._interface_name, name,
                              self._prepare_arguments(self._signals[name], args))

    def register(self) -> None:
        node_xml = f"<node name='/'><interface name='{self._interface_name}'>"
        for method_name, method_info in self._methods.items():
            node_xml += f"<method name='{method_name}'>"
            for argument in method_info[0]:
                node_xml += f"<arg type='{argument}' direction='in'/>"
            if method_info[1]:
                node_xml += f"<arg type='{method_info[1]}' direction='out'/>"
            node_xml += "</method>"
        for signal_name, signal_signature in self._signals.items():
            node_xml += f"<signal name='{signal_name}'>"
            if signal_signature:
                node_xml += f"<arg type='{signal_signature}'/>"
            node_xml += "</signal>"
        node_xml += "</interface></node>"

        node_info = Gio.DBusNodeInfo.new_for_xml(node_xml)

        regid = self._bus.register_object(
            self._path,
            node_info.interfaces[0],
            self._handle_method_call,
            None,
            None)

        if regid:
            self._regid = regid
        else:
            raise GLib.Error(f"Failed to register object with path: {self._path}")

    def unregister(self) -> None:
        if self._regid is not None:
            self._bus.unregister_object(self._regid)
            self._regid = None

    def _reregister(self) -> None:
        if self._regid:
            self.unregister()
            self.register()

    def _handle_method_call(self, _connection: Gio.DBusConnection, sender: str, _path: str, interface_name: str,
                            method_name: str, parameters: GLib.Variant, invocation: Gio.DBusMethodInvocation) -> None:
        try:
            try:
                _arguments, result_signature, method, options = self._methods[method_name]
            except KeyError:
                logging.warning(f"Unhandled method: {method_name}")
                invocation.return_error_literal(Gio.dbus_error_quark(), Gio.DBusError.UNKNOWN_METHOD,
                                                f"No such method on interface: {interface_name}.{method_name}")

            def ok(*result: Any) -> None:
                invocation.return_value(self._prepare_arguments(result_signature, result))

            args = parameters.unpack()
            if "sender" in options:
                args += (sender,)
            if "async" in options:
                method(*(args + (ok, lambda exception: self._return_dbus_error(invocation, exception))))
            else:
                ok(method(*args))
        except Exception as e:
            self._return_dbus_error(invocation, e)

    @staticmethod
    def _return_dbus_error(invocation: Gio.DBusMethodInvocation, data: object) -> None:
        if isinstance(data, DbusError):
            invocation.return_dbus_error(data.name, data.message)
        else:
            if isinstance(data, Exception):
                et, ev, etb = sys.exc_info()
                if ev is data:
                    message = "".join(traceback.format_exception(et, ev, etb))
                else:
                    message = "".join(traceback.format_exception_only(data.__class__, data))
            else:
                message = str(data)
            invocation.return_error_literal(Gio.dbus_error_quark(), Gio.DBusError.FAILED, message)

    @staticmethod
    def _prepare_arguments(signature: str, args: Collection[Any]) -> Optional[GLib.Variant]:
        return GLib.Variant(f"({signature})", args) if signature else None



from typing import Optional, Callable

from blueman.bluez.Base import Base
from blueman.bluez.AnyBase import AnyBase
from blueman.bluez.errors import BluezDBusException


class Device(Base):
    _interface_name = 'org.bluez.Device1'

    def __init__(self, obj_path: str):
        super().__init__(obj_path=obj_path)

    def pair(
        self,
        reply_handler: Optional[Callable[[], None]] = None,
        error_handler: Optional[Callable[[BluezDBusException], None]] = None,
    ) -> None:
        self._call('Pair', reply_handler=reply_handler, error_handler=error_handler)

    def connect(  # type: ignore
        self,
        reply_handler: Optional[Callable[[], None]] = None,
        error_handler: Optional[Callable[[BluezDBusException], None]] = None,
    ) -> None:
        self._call('Connect', reply_handler=reply_handler, error_handler=error_handler)

    def disconnect(  # type: ignore
        self,
        reply_handler: Optional[Callable[[], None]] = None,
        error_handler: Optional[Callable[[BluezDBusException], None]] = None,
    ) -> None:
        self._call('Disconnect', reply_handler=reply_handler, error_handler=error_handler)


class AnyDevice(AnyBase):
    def __init__(self) -> None:
        super().__init__('org.bluez.Device1')



import builtins
import typing

from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gio


class Pixbuf(GObject.Object, Gio.Icon, Gio.LoadableIcon):
    class _Props:
        @property
        def bits_per_sample(self) -> int: ...

        @property
        def has_alpha(self) -> bool: ...

        @property
        def height(self) -> int: ...

        @property
        def n_channels(self) -> int: ...

        @property
        def rowstride(self) -> int: ...

        @property
        def width(self) -> int: ...

    props: _Props

    def add_alpha(self, substitute_color: builtins.bool, r: builtins.int, g: builtins.int, b: builtins.int) -> Pixbuf: ...

    def apply_embedded_orientation(self) -> Pixbuf: ...

    @staticmethod
    def calculate_rowstride(colorspace: Colorspace, has_alpha: builtins.bool, bits_per_sample: builtins.int, width: builtins.int, height: builtins.int) -> builtins.int: ...

    def composite(self, dest: Pixbuf, dest_x: builtins.int, dest_y: builtins.int, dest_width: builtins.int, dest_height: builtins.int, offset_x: builtins.float, offset_y: builtins.float, scale_x: builtins.float, scale_y: builtins.float, interp_type: InterpType, overall_alpha: builtins.int) -> None: ...

    def composite_color(self, dest: Pixbuf, dest_x: builtins.int, dest_y: builtins.int, dest_width: builtins.int, dest_height: builtins.int, offset_x: builtins.float, offset_y: builtins.float, scale_x: builtins.float, scale_y: builtins.float, interp_type: InterpType, overall_alpha: builtins.int, check_x: builtins.int, check_y: builtins.int, check_size: builtins.int, color1: builtins.int, color2: builtins.int) -> None: ...

    def composite_color_simple(self, dest_width: builtins.int, dest_height: builtins.int, interp_type: InterpType, overall_alpha: builtins.int, check_size: builtins.int, color1: builtins.int, color2: builtins.int) -> typing.Optional[Pixbuf]: ...

    def copy(self) -> typing.Optional[Pixbuf]: ...

    def copy_area(self, src_x: builtins.int, src_y: builtins.int, width: builtins.int, height: builtins.int, dest_pixbuf: Pixbuf, dest_x: builtins.int, dest_y: builtins.int) -> None: ...

    def copy_options(self, dest_pixbuf: Pixbuf) -> builtins.bool: ...

    def fill(self, pixel: builtins.int) -> None: ...

    def flip(self, horizontal: builtins.bool) -> typing.Optional[Pixbuf]: ...

    def get_bits_per_sample(self) -> builtins.int: ...

    def get_byte_length(self) -> builtins.int: ...

    def get_colorspace(self) -> Colorspace: ...

    @staticmethod
    def get_file_info(filename: builtins.str) -> typing.Tuple[typing.Optional[PixbufFormat], builtins.int, builtins.int]: ...

    @staticmethod
    def get_file_info_async(filename: builtins.str, cancellable: typing.Optional[Gio.Cancellable], callback: typing.Optional[Gio.AsyncReadyCallback], *user_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def get_file_info_finish(async_result: Gio.AsyncResult) -> typing.Tuple[PixbufFormat, builtins.int, builtins.int]: ...

    @staticmethod
    def get_formats() -> typing.Sequence[PixbufFormat]: ...

    def get_has_alpha(self) -> builtins.bool: ...

    def get_height(self) -> builtins.int: ...

    def get_n_channels(self) -> builtins.int: ...

    def get_option(self, key: builtins.str) -> builtins.str: ...

    def get_options(self) -> typing.Mapping[builtins.str, builtins.str]: ...

    def get_pixels(self) -> builtins.bytes: ...

    def get_rowstride(self) -> builtins.int: ...

    def get_width(self) -> builtins.int: ...

    @staticmethod
    def init_modules(path: builtins.str) -> builtins.bool: ...

    @staticmethod
    def new(colorspace: Colorspace, has_alpha: builtins.bool, bits_per_sample: builtins.int, width: builtins.int, height: builtins.int, **kwargs) -> typing.Optional[Pixbuf]: ...  # type: ignore

    @staticmethod
    def new_from_bytes(data: GLib.Bytes, colorspace: Colorspace, has_alpha: builtins.bool, bits_per_sample: builtins.int, width: builtins.int, height: builtins.int, rowstride: builtins.int) -> Pixbuf: ...

    @staticmethod
    def new_from_data(data: builtins.bytes, colorspace: Colorspace, has_alpha: builtins.bool, bits_per_sample: builtins.int, width: builtins.int, height: builtins.int, rowstride: builtins.int, destroy_fn: typing.Optional[PixbufDestroyNotify], *destroy_fn_data: typing.Optional[builtins.object]) -> Pixbuf: ...

    @staticmethod
    def new_from_file(filename: builtins.str) -> Pixbuf: ...

    @staticmethod
    def new_from_file_at_scale(filename: builtins.str, width: builtins.int, height: builtins.int, preserve_aspect_ratio: builtins.bool) -> Pixbuf: ...

    @staticmethod
    def new_from_file_at_size(filename: builtins.str, width: builtins.int, height: builtins.int) -> Pixbuf: ...

    @staticmethod
    def new_from_inline(data: builtins.bytes, copy_pixels: builtins.bool) -> Pixbuf: ...

    @staticmethod
    def new_from_resource(resource_path: builtins.str) -> Pixbuf: ...

    @staticmethod
    def new_from_resource_at_scale(resource_path: builtins.str, width: builtins.int, height: builtins.int, preserve_aspect_ratio: builtins.bool) -> Pixbuf: ...

    @staticmethod
    def new_from_stream(stream: Gio.InputStream, cancellable: typing.Optional[Gio.Cancellable]) -> Pixbuf: ...

    @staticmethod
    def new_from_stream_async(stream: Gio.InputStream, cancellable: typing.Optional[Gio.Cancellable], callback: typing.Optional[Gio.AsyncReadyCallback], *user_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def new_from_stream_at_scale(stream: Gio.InputStream, width: builtins.int, height: builtins.int, preserve_aspect_ratio: builtins.bool, cancellable: typing.Optional[Gio.Cancellable]) -> Pixbuf: ...

    @staticmethod
    def new_from_stream_at_scale_async(stream: Gio.InputStream, width: builtins.int, height: builtins.int, preserve_aspect_ratio: builtins.bool, cancellable: typing.Optional[Gio.Cancellable], callback: typing.Optional[Gio.AsyncReadyCallback], *user_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def new_from_stream_finish(async_result: Gio.AsyncResult) -> Pixbuf: ...

    @staticmethod
    def new_from_xpm_data(data: typing.Sequence[builtins.str]) -> Pixbuf: ...

    def new_subpixbuf(self, src_x: builtins.int, src_y: builtins.int, width: builtins.int, height: builtins.int) -> Pixbuf: ...

    def read_pixel_bytes(self) -> GLib.Bytes: ...

    def read_pixels(self) -> builtins.int: ...

    def remove_option(self, key: builtins.str) -> builtins.bool: ...

    def rotate_simple(self, angle: PixbufRotation) -> typing.Optional[Pixbuf]: ...

    def saturate_and_pixelate(self, dest: Pixbuf, saturation: builtins.float, pixelate: builtins.bool) -> None: ...

    def save_to_bufferv(self, type: builtins.str, option_keys: typing.Sequence[builtins.str], option_values: typing.Sequence[builtins.str]) -> typing.Tuple[builtins.bool, builtins.bytes]: ...

    def save_to_callbackv(self, save_func: PixbufSaveFunc, user_data: typing.Optional[builtins.object], type: builtins.str, option_keys: typing.Sequence[builtins.str], option_values: typing.Sequence[builtins.str]) -> builtins.bool: ...

    @staticmethod
    def save_to_stream_finish(async_result: Gio.AsyncResult) -> builtins.bool: ...

    def save_to_streamv(self, stream: Gio.OutputStream, type: builtins.str, option_keys: typing.Sequence[builtins.str], option_values: typing.Sequence[builtins.str], cancellable: typing.Optional[Gio.Cancellable]) -> builtins.bool: ...

    def save_to_streamv_async(self, stream: Gio.OutputStream, type: builtins.str, option_keys: typing.Sequence[builtins.str], option_values: typing.Sequence[builtins.str], cancellable: typing.Optional[Gio.Cancellable], callback: typing.Optional[Gio.AsyncReadyCallback], *user_data: typing.Optional[builtins.object]) -> None: ...

    def savev(self, filename: builtins.str, type: builtins.str, option_keys: typing.Sequence[builtins.str], option_values: typing.Sequence[builtins.str]) -> builtins.bool: ...

    def scale(self, dest: Pixbuf, dest_x: builtins.int, dest_y: builtins.int, dest_width: builtins.int, dest_height: builtins.int, offset_x: builtins.float, offset_y: builtins.float, scale_x: builtins.float, scale_y: builtins.float, interp_type: InterpType) -> None: ...

    def scale_simple(self, dest_width: builtins.int, dest_height: builtins.int, interp_type: InterpType) -> typing.Optional[Pixbuf]: ...

    def set_option(self, key: builtins.str, value: builtins.str) -> builtins.bool: ...


class PixbufAnimation(GObject.Object):

    def get_height(self) -> builtins.int: ...

    def get_iter(self, start_time: typing.Optional[GLib.TimeVal]) -> PixbufAnimationIter: ...

    def get_static_image(self) -> Pixbuf: ...

    def get_width(self) -> builtins.int: ...

    def is_static_image(self) -> builtins.bool: ...

    @staticmethod
    def new_from_file(filename: builtins.str) -> PixbufAnimation: ...

    @staticmethod
    def new_from_resource(resource_path: builtins.str) -> PixbufAnimation: ...

    @staticmethod
    def new_from_stream(stream: Gio.InputStream, cancellable: typing.Optional[Gio.Cancellable]) -> PixbufAnimation: ...

    @staticmethod
    def new_from_stream_async(stream: Gio.InputStream, cancellable: typing.Optional[Gio.Cancellable], callback: typing.Optional[Gio.AsyncReadyCallback], *user_data: typing.Optional[builtins.object]) -> None: ...

    @staticmethod
    def new_from_stream_finish(async_result: Gio.AsyncResult) -> PixbufAnimation: ...


class PixbufAnimationIter(GObject.Object):

    def advance(self, current_time: typing.Optional[GLib.TimeVal]) -> builtins.bool: ...

    def get_delay_time(self) -> builtins.int: ...

    def get_pixbuf(self) -> Pixbuf: ...

    def on_currently_loading_frame(self) -> builtins.bool: ...


class PixbufLoader(GObject.Object):
    parent_instance: GObject.Object
    priv: builtins.object

    def close(self) -> builtins.bool: ...

    def get_animation(self) -> PixbufAnimation: ...

    def get_format(self) -> typing.Optional[PixbufFormat]: ...

    def get_pixbuf(self) -> Pixbuf: ...

    @staticmethod
    def new(**kwargs) -> PixbufLoader: ...  # type: ignore

    @staticmethod
    def new_with_mime_type(mime_type: builtins.str) -> PixbufLoader: ...

    @staticmethod
    def new_with_type(image_type: builtins.str) -> PixbufLoader: ...

    def set_size(self, width: builtins.int, height: builtins.int) -> None: ...

    def write(self, buf: builtins.bytes) -> builtins.bool: ...

    def write_bytes(self, buffer: GLib.Bytes) -> builtins.bool: ...

    def do_area_prepared(self) -> None: ...

    def do_area_updated(self, x: builtins.int, y: builtins.int, width: builtins.int, height: builtins.int) -> None: ...

    def do_closed(self) -> None: ...

    def do_size_prepared(self, width: builtins.int, height: builtins.int) -> None: ...


class PixbufSimpleAnim(PixbufAnimation):

    def add_frame(self, pixbuf: Pixbuf) -> None: ...

    def get_loop(self) -> builtins.bool: ...

    @staticmethod
    def new(width: builtins.int, height: builtins.int, rate: builtins.float) -> PixbufSimpleAnim: ...

    def set_loop(self, loop: builtins.bool) -> None: ...


class PixbufSimpleAnimIter(PixbufAnimationIter):
    ...


class PixbufFormat():

    def copy(self) -> PixbufFormat: ...

    def free(self) -> None: ...

    def get_description(self) -> builtins.str: ...

    def get_extensions(self) -> typing.Sequence[builtins.str]: ...

    def get_license(self) -> builtins.str: ...

    def get_mime_types(self) -> typing.Sequence[builtins.str]: ...

    def get_name(self) -> builtins.str: ...

    def is_disabled(self) -> builtins.bool: ...

    def is_save_option_supported(self, option_key: builtins.str) -> builtins.bool: ...

    def is_scalable(self) -> builtins.bool: ...

    def is_writable(self) -> builtins.bool: ...

    def set_disabled(self, disabled: builtins.bool) -> None: ...


class Colorspace(GObject.GEnum, builtins.int):
    RGB = ...  # type: Colorspace


class InterpType(GObject.GEnum, builtins.int):
    BILINEAR = ...  # type: InterpType
    HYPER = ...  # type: InterpType
    NEAREST = ...  # type: InterpType
    TILES = ...  # type: InterpType


class PixbufAlphaMode(GObject.GEnum, builtins.int):
    BILEVEL = ...  # type: PixbufAlphaMode
    FULL = ...  # type: PixbufAlphaMode


class PixbufError(GObject.GEnum, builtins.int):
    BAD_OPTION = ...  # type: PixbufError
    CORRUPT_IMAGE = ...  # type: PixbufError
    FAILED = ...  # type: PixbufError
    INCOMPLETE_ANIMATION = ...  # type: PixbufError
    INSUFFICIENT_MEMORY = ...  # type: PixbufError
    UNKNOWN_TYPE = ...  # type: PixbufError
    UNSUPPORTED_OPERATION = ...  # type: PixbufError

    @staticmethod
    def quark() -> builtins.int: ...


class PixbufRotation(GObject.GEnum, builtins.int):
    CLOCKWISE = ...  # type: PixbufRotation
    COUNTERCLOCKWISE = ...  # type: PixbufRotation
    NONE = ...  # type: PixbufRotation
    UPSIDEDOWN = ...  # type: PixbufRotation


PixbufDestroyNotify = typing.Callable[[builtins.bytes, typing.Optional[builtins.object]], None]
PixbufSaveFunc = typing.Callable[[builtins.bytes, typing.Optional[builtins.object]], typing.Tuple[builtins.bool, GLib.Error]]


def pixbuf_error_quark() -> builtins.int: ...


PIXBUF_FEATURES_H: builtins.int
PIXBUF_MAJOR: builtins.int
PIXBUF_MICRO: builtins.int
PIXBUF_MINOR: builtins.int
PIXBUF_VERSION: builtins.str



from typing import Dict, List, Optional

from gi.repository import GObject, GLib
from gi.repository import Gio

from blueman.bluemantyping import GSignals


class AnyBase(GObject.GObject):
    __gsignals__: GSignals = {
        'property-changed': (GObject.SignalFlags.NO_HOOKS, None, (str, object, str))
    }

    connect_signal = GObject.GObject.connect
    disconnect_signal = GObject.GObject.disconnect

    __bus_name = 'org.bluez'
    __bus_interface_name = 'org.freedesktop.DBus.Properties'

    def __init__(self, interface_name: str):
        super().__init__()

        self.__bus: Optional[Gio.DBusConnection] = Gio.bus_get_sync(Gio.BusType.SYSTEM)

        self.__interface_name = interface_name
        self.__signal = None

        def on_signal(
            _connection: Gio.DBusConnection,
            _sender_name: str,
            object_path: str,
            _interface_name: str,
            _signal_name: str,
            param: GLib.Variant,
        ) -> None:
            iface_name, changed, invalidated = param.unpack()
            if iface_name == self.__interface_name:
                self._on_properties_changed(object_path, changed, invalidated)

        self.__signal = self.__bus.signal_subscribe(
            self.__bus_name, self.__bus_interface_name, 'PropertiesChanged', None, None,
            Gio.DBusSignalFlags.NONE, on_signal)

    def _on_properties_changed(
        self, object_path: str, changed_properties: Dict[str, object], _invalidated: List[str]
    ) -> None:
        for name, value in changed_properties.items():
            self.emit('property-changed', name, value, object_path)

    def close(self) -> None:
        if self.__signal:
            if self.__bus is not None:
                self.__bus.signal_unsubscribe(self.__signal)
            self.__signal = None
        self.__bus = None



from typing import Dict, Tuple, Union, TYPE_CHECKING
from gi.repository import GObject

if TYPE_CHECKING:
    from typing_extensions import Protocol

    class _HasGType(Protocol):
        __gtype__: GObject.GType

# Actually supported types are int, bool, str, float, and object but no subclasses, see
# https://github.com/GNOME/pygobject/blob/ac576400ecd554879c906791e6638d64bb8bcc2a/gi/pygi-type.c#L498
# (We shield the possibility to provide a str to avoid errors)
GSignals = Dict[str, Tuple[GObject.SignalFlags, None, Tuple[Union[None, type, GObject.GType, "_HasGType"], ...]]]



from gettext import gettext as _
from typing import Set

from blueman.Service import Action
from blueman.services.meta import SerialService
from blueman.Sdp import DIALUP_NET_SVCLASS_ID
from blueman.gui.GsmSettings import GsmSettings


class DialupNetwork(SerialService):
    __svclass_id__ = DIALUP_NET_SVCLASS_ID
    __icon__ = "modem"
    __priority__ = 50

    @property
    def common_actions(self) -> Set[Action]:
        def open_settings() -> None:
            d = GsmSettings(self.device['Address'])
            d.run()
            d.destroy()

        return {Action(
            _("Dialup Settings"),
            "preferences-other",
            {'PPPSupport', 'NMDUNSupport'},
            open_settings
        )}



from gettext import gettext as _
from typing import TYPE_CHECKING, Callable, Union

from _blueman import RFCOMMError
from blueman.services.meta.SerialService import SerialService

from blueman.bluez.Device import Device
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.gui.Notification import Notification
from blueman.main.DBusProxies import Mechanism
from blueman.main.Config import Config

from gi.repository import GLib

import subprocess
import logging

from blueman.plugins.applet.DBusService import RFCOMMConnectHandler
from blueman.services import DialupNetwork

if TYPE_CHECKING:
    from blueman.main.Applet import BluemanApplet


class PPPConnectedListener:
    def on_ppp_connected(self, device: Device, rfcomm: str, ppp_port: str) -> None:
        ...


class Connection:
    def __init__(self, applet: "BluemanApplet", service: DialupNetwork, port: str,
                 ok: Callable[[str], None], err: Callable[[GLib.Error], None]):
        self.reply_handler = ok
        self.error_handler = err
        self.service = service
        self.port = port
        self.parent = applet

        stdout, stderr = subprocess.Popen(['ps', 'ax', '-o', 'pid,args'], stdout=subprocess.PIPE).communicate()
        if b'ModemManager' in stdout:
            timeout = 10
            logging.info(f"ModemManager is running, delaying connection {timeout} sec for it to complete probing")
            GLib.timeout_add_seconds(timeout, self.connect)
        else:
            self.connect()

    def connect(self) -> bool:
        c = Config("org.blueman.gsmsetting", f"/org/blueman/gsmsettings/{self.service.device['Address']}/")

        m = Mechanism()
        m.PPPConnect('(sss)', self.port, c["number"], c["apn"], result_handler=self.on_connected,
                     error_handler=self.on_error)

        return False

    def on_error(self, _obj: Mechanism, result: GLib.Error, _user_data: None) -> None:
        logging.info(f"Failed {result}")
        # FIXME confusingly self.port is the full rfcomm device path but the service expects the number only
        self.error_handler(result)

        def _connect() -> bool:
            self.service.disconnect(int(self.port[-1]))
            return False

        GLib.timeout_add(1000, _connect)

    def on_connected(self, _obj: Mechanism, result: str, _user_data: None) -> None:
        self.reply_handler(self.port)
        for plugin in self.parent.Plugins.get_loaded_plugins(PPPConnectedListener):
            plugin.on_ppp_connected(self.service.device, self.port, result)

        msg = _("Successfully connected to <b>DUN</b> service on <b>%(0)s.</b>\n"
                "Network is now available through <b>%(1)s</b>") % {"0": self.service.device['Alias'], "1": result}

        Notification(_("Connected"), msg, icon_name="network-wireless").show()


class PPPSupport(AppletPlugin, RFCOMMConnectHandler):
    __depends__ = ["DBusService"]
    __description__ = _("Provides basic support for connecting to the internet via DUN profile.")
    __author__ = "Walmis"
    __icon__ = "modem"
    __priority__ = 0

    def rfcomm_connect_handler(self, service: SerialService, reply: Callable[[str], None],
                               err: Callable[[Union[RFCOMMError, GLib.Error]], None]) -> bool:
        if isinstance(service, DialupNetwork):
            def local_reply(port: str) -> None:
                assert isinstance(service, DialupNetwork)  # https://github.com/python/mypy/issues/2608
                Connection(self.parent, service, port, reply, err)

            service.connect(reply_handler=local_reply, error_handler=err)
            logging.info("Connecting rfcomm device")

            return True
        else:
            return False
